/*
    CHAOS target for netfilter

    Copyright © Jan Engelhardt <jengelh [at] gmx de>, 2006 - 2007
    released under the terms of the GNU General Public
    License version 2.x and only versions 2.x.
*/
#include <linux/icmp.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/stat.h>
#include <linux/netfilter/x_tables.h>
#include <linux/netfilter/xt_tcpudp.h>
#include <linux/netfilter_ipv4/ipt_REJECT.h>
#include <net/ip.h>
#include <linux/netfilter/xt_CHAOS.h>
#include "oot_trans.h"
#define PFX KBUILD_MODNAME ": "

/* Module parameters */
static unsigned int reject_percentage = ~0U * .01;
static unsigned int delude_percentage = ~0U * .0101;
module_param(reject_percentage, uint, S_IRUGO | S_IWUSR);
module_param(delude_percentage, uint, S_IRUGO | S_IWUSR);

/* References to other matches/targets */
static struct xt_match *xm_tcp;
static struct xt_target *xt_delude, *xt_reject, *xt_tarpit;

static int have_delude, have_tarpit;

/* Static data for other matches/targets */
static const struct ipt_reject_info reject_params = {
	.with = ICMP_HOST_UNREACH,
};

static const struct xt_tcp tcp_params = {
	.spts = {0, ~0},
	.dpts = {0, ~0},
};

/* CHAOS functions */
static void xt_chaos_total(const struct xt_chaos_info *info,
    struct sk_buff **pskb, const struct net_device *in,
    const struct net_device *out, unsigned int hooknum)
{
	const int protoff = 4 * (*pskb)->nh.iph->ihl;
	const int offset  = ntohs((*pskb)->nh.iph->frag_off) & IP_OFFSET;
	const struct xt_target *destiny;
	int hotdrop = 0, ret;

	ret = xm_tcp->match(*pskb, in, out, xm_tcp, &tcp_params,
	                    offset, protoff, &hotdrop);
	if(!ret || hotdrop || (unsigned int)net_random() > delude_percentage)
		return;

	destiny = (info->variant == XTCHAOS_TARPIT) ? xt_tarpit : xt_delude;
#ifdef HAVE_TARGUSERINFO
	destiny->target(pskb, in, out, hooknum, destiny, NULL, NULL);
#else
	destiny->target(pskb, in, out, hooknum, destiny, NULL);
#endif
	return;
}

static unsigned int xt_chaos_target(struct sk_buff **pskb,
    const struct net_device *in, const struct net_device *out,
    unsigned int hooknum, const struct xt_target *target, const void *targinfo
#ifdef HAVE_TARGUSERINFO
    ,
    void *userinfo
#endif
    )
{
	/* Equivalent to:
	 * -A chaos -m statistic --mode random --probability \
	 *         $reject_percentage -j REJECT --reject-with host-unreach;
	 * -A chaos -m statistic --mode random --probability \
	 *         $delude_percentage -j DELUDE;
	 * -A chaos -j DROP;
	 */
	const struct xt_chaos_info *info = targinfo;

	if((unsigned int)net_random() <= reject_percentage)
#ifdef HAVE_TARGUSERINFO
		return xt_reject->target(pskb, in, out, hooknum, target,
		       &reject_params, userinfo);
#else
		return xt_reject->target(pskb, in, out, hooknum, target,
		       &reject_params);
#endif

	/* TARPIT/DELUDE may not be called from the OUTPUT chain */
	if((*pskb)->nh.iph->protocol == IPPROTO_TCP &&
	  info->variant != XTCHAOS_NORMAL && hooknum != NF_IP_LOCAL_OUT)
		xt_chaos_total(info, pskb, in, out, hooknum);

	return NF_DROP;
}

static int xt_chaos_checkentry(const char *tablename, const void *entry,
    const struct xt_target *target, void *targinfo,
#ifdef HAVE_TARGINFOSIZE
    unsigned int targinfosize,
#endif
    unsigned int hook_mask)
{
	const struct xt_chaos_info *info = targinfo;
	if(info->variant == XTCHAOS_DELUDE && !have_delude) {
		printk(KERN_WARNING PFX "Error: Cannot use --delude when "
		       "DELUDE module not available\n");
		return 0;
	}
	if(info->variant == XTCHAOS_TARPIT && !have_tarpit) {
		printk(KERN_WARNING PFX "Error: Cannot use --tarpit when "
		       "TARPIT module not available\n");
		return 0;
	}
	return 1;
}

static struct xt_target xt_chaos_info = {
	.name       = "CHAOS",
	.target     = xt_chaos_target,
	.checkentry = xt_chaos_checkentry,
	.table      = "filter",
	.targetsize = sizeof(struct xt_chaos_info),
	.hooks      = (1 << NF_IP_LOCAL_IN) | (1 << NF_IP_FORWARD) |
	              (1 << NF_IP_LOCAL_OUT),
	.family     = AF_INET,
	.me         = THIS_MODULE,
};

static int __init xt_chaos_init(void)
{
	int ret = -EINVAL;

	xm_tcp = xt_request_find_match(AF_INET, "tcp", 0);
	if(xm_tcp == NULL) {
		printk(KERN_WARNING PFX "Error: Could not find or load "
		       "\"tcp\" match\n");
		return -EINVAL;
	}

	xt_reject = xt_request_find_target(AF_INET, "REJECT", 0);
	if(xt_reject == NULL) {
		printk(KERN_WARNING PFX "Error: Could not find or load "
		       "\"REJECT\" target\n");
		goto out2;
	}

	xt_tarpit   = xt_request_find_target(AF_INET, "TARPIT", 0);
	have_tarpit = xt_tarpit != NULL;
	if(!have_tarpit)
		printk(KERN_WARNING PFX "Warning: Could not find or load "
		       "\"TARPIT\" target\n");

	xt_delude   = xt_request_find_target(AF_INET, "DELUDE", 0);
	have_delude = xt_delude != NULL;
	if(!have_delude)
		printk(KERN_WARNING PFX "Warning: Could not find or load "
		       "\"DELUDE\" target\n");

	if((ret = xt_register_target(&xt_chaos_info)) != 0) {
		printk(KERN_WARNING PFX "xt_register_target returned "
		       "error %d\n", ret);
		goto out3;
	}

	return 0;

 out3:
 	if(have_delude)
 		module_put(xt_delude->me);
	if(have_tarpit)
		module_put(xt_tarpit->me);
	module_put(xt_reject->me);
 out2:
	module_put(xm_tcp->me);
	return ret;
}

static void __exit xt_chaos_exit(void)
{
	xt_unregister_target(&xt_chaos_info);
	module_put(xm_tcp->me);
	module_put(xt_reject->me);
	if(have_delude)
		module_put(xt_delude->me);
	if(have_tarpit)
		module_put(xt_tarpit->me);
	return;
}

module_init(xt_chaos_init);
module_exit(xt_chaos_exit);
MODULE_AUTHOR("Jan Engelhardt <jengelh@gmx.de>");
MODULE_DESCRIPTION("netfilter CHAOS target");
MODULE_LICENSE("GPL");
MODULE_ALIAS("ipt_CHAOS");
