/*
    portscan match for netfilter

    Written by Jan Engelhardt, 2006 - 2007
    released under the terms of the GNU General Public
    License version 2.x and only versions 2.x.
*/
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/skbuff.h>
#include <linux/stat.h>
#include <linux/tcp.h>
#include <linux/types.h>
#include <linux/version.h>
#include <linux/netfilter/x_tables.h>
#include <linux/netfilter/xt_tcpudp.h>
#include "oot_conntrack.h"
#include <linux/netfilter/xt_portscan.h>
#include "oot_trans.h"
#define PFX KBUILD_MODNAME ": "

enum {
	TCP_FLAGS_ALL3 = TCP_FLAG_FIN | TCP_FLAG_RST | TCP_FLAG_SYN,
	TCP_FLAGS_ALL4 = TCP_FLAGS_ALL3 | TCP_FLAG_ACK,
	TCP_FLAGS_ALL6 = TCP_FLAGS_ALL4 | TCP_FLAG_PSH | TCP_FLAG_URG,
};

/* Module parameters */
static unsigned int
	connmark_mask = ~0,
	packet_mask   = ~0,
	mark_seen     = 0x9,
	mark_synrcv   = 0x1,
	mark_closed   = 0x2,
	mark_synscan  = 0x3,
	mark_estab1   = 0x4,
	mark_estab2   = 0x5,
	mark_cnscan   = 0x6,
	mark_grscan   = 0x7,
	mark_valid    = 0x8;

module_param(connmark_mask, uint, S_IRUGO | S_IWUSR);
module_param(packet_mask,   uint, S_IRUGO | S_IWUSR);
module_param(mark_seen,     uint, S_IRUGO | S_IWUSR);
module_param(mark_synrcv,   uint, S_IRUGO | S_IWUSR);
module_param(mark_closed,   uint, S_IRUGO | S_IWUSR);
module_param(mark_synscan,  uint, S_IRUGO | S_IWUSR);
module_param(mark_estab1,   uint, S_IRUGO | S_IWUSR);
module_param(mark_estab2,   uint, S_IRUGO | S_IWUSR);
module_param(mark_cnscan,   uint, S_IRUGO | S_IWUSR);
module_param(mark_grscan,   uint, S_IRUGO | S_IWUSR);
module_param(mark_valid,    uint, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(connmark_mask, "only set specified bits in connection mark");
MODULE_PARM_DESC(packet_mask,   "only set specified bits in packet mark");
MODULE_PARM_DESC(mark_seen,     "nfmark value for packet-seen state");
MODULE_PARM_DESC(mark_synrcv,   "connmark value for SYN Received state");
MODULE_PARM_DESC(mark_closed,   "connmark value for closed state");
MODULE_PARM_DESC(mark_synscan,  "connmark value for SYN Scan state");
MODULE_PARM_DESC(mark_estab1,   "connmark value for Established-1 state");
MODULE_PARM_DESC(mark_estab2,   "connmark value for Established-2 state");
MODULE_PARM_DESC(mark_cnscan,   "connmark value for Connect Scan state");
MODULE_PARM_DESC(mark_grscan,   "connmark value for Grab Scan state");
MODULE_PARM_DESC(mark_valid,    "connmark value for Valid state");

/* TCP flag functions */
static inline int tflg_ack4(const struct tcphdr *th)
{
	return (tcp_flag_word(th) & TCP_FLAGS_ALL4) == TCP_FLAG_ACK;
}

static inline int tflg_ack6(const struct tcphdr *th)
{
	return (tcp_flag_word(th) & TCP_FLAGS_ALL6) == TCP_FLAG_ACK;
}

static inline int tflg_fin(const struct tcphdr *th)
{
	return (tcp_flag_word(th) & TCP_FLAGS_ALL3) == TCP_FLAG_FIN;
}

static inline int tflg_rst(const struct tcphdr *th)
{
	return (tcp_flag_word(th) & TCP_FLAGS_ALL3) == TCP_FLAG_RST;
}

static inline int tflg_rstack(const struct tcphdr *th)
{
	return (tcp_flag_word(th) & TCP_FLAGS_ALL4) ==
	       (TCP_FLAG_ACK | TCP_FLAG_RST);
}

static inline int tflg_syn(const struct tcphdr *th)
{
	return (tcp_flag_word(th) & TCP_FLAGS_ALL4) == TCP_FLAG_SYN;
}

static inline int tflg_synack(const struct tcphdr *th)
{
	return (tcp_flag_word(th) & TCP_FLAGS_ALL4) ==
	       (TCP_FLAG_SYN | TCP_FLAG_ACK);
}

/* portscan functions */
static inline int xt_portscan_stealth(const struct tcphdr *th)
{
	/*
	 * "Connection refused" replies to our own probes must not be matched.
	 */
	if(tflg_rstack(th))
		return 0;

	if(tflg_rst(th) && printk_ratelimit()) {
		printk(KERN_WARNING PFX "Warning: Pure RST received\n");
		return 0;
	}

	/*
	 * -p tcp ! --syn -m conntrack --ctstate INVALID: Looking for non-start
	 * packets that are not associated with any connection -- this will
	 * match most scan types (NULL, XMAS, FIN) and ridiculous flag
	 * combinations (SYN-RST, SYN-FIN, SYN-FIN-RST, FIN-RST, etc.).
	 */
	return !tflg_syn(th);
}

static inline int xt_portscan_full(int mark, enum ip_conntrack_info ctstate,
    int loopback, const struct tcphdr *tcph, int payload_len)
{
	if(mark == mark_estab2) {
		/*
		 * -m connmark --mark $ESTAB2
		 */
		if(tflg_ack4(tcph) && payload_len == 0)
			return mark; /* keep mark */
		else if(tflg_rst(tcph) || tflg_fin(tcph))
			return mark_grscan;
		else
			return mark_valid;
	} else if(mark == mark_estab1) {
		/*
		 * -m connmark --mark $ESTAB1
		 */
		if(tflg_rst(tcph) || tflg_fin(tcph))
			return mark_cnscan;
		else if(!loopback && tflg_ack4(tcph) && payload_len == 0)
			return mark_estab2;
		else
			return mark_valid;
	} else if(mark == mark_synrcv) {
		/*
		 * -m connmark --mark $SYN
		 */
		if(loopback && tflg_synack(tcph))
			return mark; /* keep mark */
		else if(loopback && tflg_rstack(tcph))
			return mark_closed;
		else if(tflg_ack6(tcph))
			return mark_estab1;
		else
			return mark_synscan;
	} else if(ctstate == IP_CT_NEW && tflg_syn(tcph)) {
		/*
		 * -p tcp --syn --ctstate NEW
		 */
		return mark_synrcv;
	}
	return mark;
}

static int xt_portscan_match(const struct sk_buff *skb,
    const struct net_device *in, const struct net_device *out,
    const struct xt_match *match, const void *matchinfo, int offset,
    unsigned int protoff, int *hotdrop)
{
	const struct xt_portscan_info *info = matchinfo;
	enum ip_conntrack_info ctstate;
	struct ip_conntrack *ctdata;
	const struct tcphdr *tcph;
	struct tcphdr tcph_buf;

	tcph = skb_header_pointer(skb, protoff, sizeof(tcph_buf), &tcph_buf);
	if(tcph == NULL)
		return 0;

	/* Check for invalid packets: -m conntrack --ctstate INVALID */
	if((ctdata = ip_conntrack_get(skb, &ctstate)) == NULL) {
		if(info->match_stealth)
			return xt_portscan_stealth(tcph);
		/*
		 * If @ctdata is NULL, we cannot match the other scan
		 * types, return.
		 */
		return 0;
	}

	/*
	 * If -m portscan was previously applied to this packet, the rules we
	 * simulate must not be run through again. And for speedup, do not call
	 * it either when the connection is already VALID.
	 */
	if((ctdata->mark & connmark_mask) == mark_valid ||
	  (skb->nfmark & packet_mask) != mark_seen)
	{
		unsigned int n;
		n = xt_portscan_full(ctdata->mark & connmark_mask, ctstate,
		    in == &loopback_dev, tcph,
		    skb->len - protoff - 4 * tcph->doff);

		ctdata->mark = (ctdata->mark & ~connmark_mask) | n;
		((struct sk_buff *)skb)->nfmark =
			(skb->nfmark & ~packet_mask) | mark_seen;
	}

	return (info->match_syn && ctdata->mark == mark_synscan) ||
	       (info->match_cn && ctdata->mark == mark_cnscan) ||
	       (info->match_gr && ctdata->mark == mark_grscan);
}

static int xt_portscan_checkentry(const char *tablename, const void *entry,
    const struct xt_match *match, void *matchinfo,
#ifdef HAVE_MATCHINFOSIZE
    unsigned int matchinfosize,
#endif
    unsigned int hook_mask)
{
	const struct xt_portscan_info *info = matchinfo;
#ifdef HAVE_MATCHINFOSIZE
	if(matchinfosize != XT_ALIGN(sizeof(struct xt_portscan_info))) {
		printk(KERN_WARNING PFX "matchinfosize %u != %Zu\n",
		       matchinfosize,
		       XT_ALIGN(sizeof(struct xt_portscan_info)));
		return 0;
	}
#endif
	if((info->match_stealth & ~1) || (info->match_syn & ~1) ||
	  (info->match_cn & ~1) || (info->match_gr & ~1)) {
		printk(KERN_WARNING PFX "Invalid flags\n");
		return 0;
	}
	return 1;
}

static struct xt_match xt_portscan = {
	.name       = "portscan",
	.match      = xt_portscan_match,
	.checkentry = xt_portscan_checkentry,
	.matchsize  = sizeof(struct xt_portscan_info),
	.proto      = IPPROTO_TCP,
	.family     = AF_INET,
	.me         = THIS_MODULE,
};

static int __init xt_portscan_init(void)
{
	return xt_register_match(&xt_portscan);
}

static void __exit xt_portscan_exit(void)
{
	xt_unregister_match(&xt_portscan);
	return;
}

module_init(xt_portscan_init);
module_exit(xt_portscan_exit);
MODULE_AUTHOR("Jan Engelhardt <jengelh@gmx.de>");
MODULE_DESCRIPTION("netfilter portscan match module");
MODULE_LICENSE("GPL");
MODULE_ALIAS("ipt_portscan");
